package net.hostsharing.hsadminng.hs.office.debitor;

import lombok.*;
import net.hostsharing.hsadminng.errors.DisplayName;
import net.hostsharing.hsadminng.hs.office.bankaccount.HsOfficeBankAccountEntity;
import net.hostsharing.hsadminng.hs.office.relationship.HsOfficeRelationshipEntity;
import net.hostsharing.hsadminng.hs.office.partner.HsOfficePartnerEntity;
import net.hostsharing.hsadminng.persistence.HasUuid;
import net.hostsharing.hsadminng.rbac.rbacdef.RbacView;
import net.hostsharing.hsadminng.rbac.rbacdef.RbacView.SQL;
import net.hostsharing.hsadminng.stringify.Stringify;
import net.hostsharing.hsadminng.stringify.Stringifyable;
import org.hibernate.annotations.GenericGenerator;
import org.hibernate.annotations.JoinFormula;
import org.hibernate.annotations.NotFound;
import org.hibernate.annotations.NotFoundAction;

import jakarta.persistence.*;
import java.io.IOException;
import java.util.UUID;

import static java.util.Optional.ofNullable;
import static net.hostsharing.hsadminng.rbac.rbacdef.RbacView.Column.dependsOnColumn;
import static net.hostsharing.hsadminng.rbac.rbacdef.RbacView.Nullable.NULLABLE;
import static net.hostsharing.hsadminng.rbac.rbacdef.RbacView.Permission.*;
import static net.hostsharing.hsadminng.rbac.rbacdef.RbacView.Role.*;
import static net.hostsharing.hsadminng.rbac.rbacdef.RbacView.SQL.fetchedBySql;
import static net.hostsharing.hsadminng.rbac.rbacdef.RbacView.rbacViewFor;
import static net.hostsharing.hsadminng.stringify.Stringify.stringify;

@Entity
@Table(name = "hs_office_debitor_rv")
@Getter
@Setter
@Builder(toBuilder = true)
@NoArgsConstructor
@AllArgsConstructor
@DisplayName("Debitor")
public class HsOfficeDebitorEntity implements HasUuid, Stringifyable {

    public static final String DEBITOR_NUMBER_TAG = "D-";

    private static Stringify<HsOfficeDebitorEntity> stringify =
            stringify(HsOfficeDebitorEntity.class, "debitor")
                    .withIdProp(HsOfficeDebitorEntity::toShortString)
                    .withProp(e -> ofNullable(e.getDebitorRel()).map(HsOfficeRelationshipEntity::toShortString).orElse(null))
                    .withProp(HsOfficeDebitorEntity::getDefaultPrefix)
                    .quotedValues(false);

    @Id
    @GeneratedValue
    @GenericGenerator(name = "UUID", strategy = "org.hibernate.id.UUIDGenerator")
    private UUID uuid;

    @ManyToOne
    @JoinFormula(
        referencedColumnName = "uuid",
        // FIXME: use _rv in sub-query
        value = """
            (
                SELECT DISTINCT partner.uuid
                FROM hs_office_partner partner
                JOIN hs_office_relationship dRel
                    ON dRel.uuid = debitorreluuid AND dRel.relType = 'ACCOUNTING'
                JOIN hs_office_relationship pRel
                    ON pRel.uuid = partner.partnerRoleUuid AND pRel.relType = 'PARTNER'
                WHERE pRel.relHolderUuid = dRel.relAnchorUuid
            )
            """)
    @NotFound(action = NotFoundAction.IGNORE)
    private HsOfficePartnerEntity partner;

    @Column(name = "debitornumbersuffix", columnDefinition = "numeric(2)")
    private Byte debitorNumberSuffix; // TODO maybe rather as a formatted String?

    @ManyToOne(cascade = CascadeType.ALL)
    @JoinColumn(name = "debitorreluuid", nullable = false)
    private HsOfficeRelationshipEntity debitorRel;

    @Column(name = "billable", nullable = false)
    private Boolean billable; // not a primitive because otherwise the default would be false

    @Column(name = "vatid")
    private String vatId;

    @Column(name = "vatcountrycode")
    private String vatCountryCode;

    @Column(name = "vatbusiness")
    private boolean vatBusiness;

    @Column(name = "vatreversecharge")
    private boolean vatReverseCharge;

    @ManyToOne
    @JoinColumn(name = "refundbankaccountuuid")
    private HsOfficeBankAccountEntity refundBankAccount;

    @Column(name = "defaultprefix", columnDefinition = "char(3) not null")
    private String defaultPrefix;

    private String getDebitorNumberString() {
        return ofNullable(partner)
                .filter(partner -> debitorNumberSuffix != null)
                .map(HsOfficePartnerEntity::getPartnerNumber)
                .map(Object::toString)
                .map(partnerNumber -> partnerNumber + String.format("%02d", debitorNumberSuffix))
                .orElse(null);
    }

    public Integer getDebitorNumber() {
        return ofNullable(getDebitorNumberString()).map(Integer::parseInt).orElse(null);
    }

    @Override
    public String toString() {
        return stringify.apply(this);
    }

    @Override
    public String toShortString() {
        return DEBITOR_NUMBER_TAG + getDebitorNumberString();
    }

    public static RbacView rbac() {
        return rbacViewFor("debitor", HsOfficeDebitorEntity.class)
                .withIdentityView(SQL.query("""
                        SELECT debitor.uuid AS uuid,
                                    'D-' || (SELECT partner.partnerNumber
                                            FROM hs_office_partner partner
                                            JOIN hs_office_relationship partnerRel
                                                ON partnerRel.uuid = partner.partnerRoleUUid AND partnerRel.relType = 'PARTNER'
                                            JOIN hs_office_relationship debitorRel
                                                ON debitorRel.relAnchorUuid = partnerRel.relHolderUuid AND debitorRel.relType = 'ACCOUNTING'
                                            WHERE debitorRel.uuid = debitor.debitorRelUuid)
                                         || to_char(debitorNumberSuffix, 'fm00') as idName
                        FROM hs_office_debitor AS debitor
                        """))
                .withRestrictedViewOrderBy(SQL.projection("defaultPrefix"))
                .withUpdatableColumns(
                        "debitorRelUuid",
                        "billable",
                        "refundBankAccountUuid",
                        "vatId",
                        "vatCountryCode",
                        "vatBusiness",
                        "vatReverseCharge",
                        "defaultPrefix" /* TODO: do we want that updatable? */)
                .toRole("global", ADMIN).grantPermission(INSERT)

                .importRootEntityAliasProxy("debitorRel", HsOfficeRelationshipEntity.class,
                        fetchedBySql("""
                                SELECT *
                                    FROM hs_office_relationship AS r
                                    WHERE r.relType = 'ACCOUNTING' AND r.uuid = ${REF}.debitorRelUuid
                                """),
                        dependsOnColumn("debitorRelUuid"))
                .createPermission(DELETE).grantedTo("debitorRel", OWNER)
                .createPermission(UPDATE).grantedTo("debitorRel", ADMIN)
                .createPermission(SELECT).grantedTo("debitorRel", TENANT)

                .importEntityAlias("refundBankAccount", HsOfficeBankAccountEntity.class,
                        dependsOnColumn("refundBankAccountUuid"),
                        fetchedBySql("""
                                SELECT *
                                    FROM hs_office_bankaccount AS b
                                    WHERE b.uuid = ${REF}.refundBankAccountUuid
                                """),
                        NULLABLE
                )
                .toRole("refundBankAccount", ADMIN).grantRole("debitorRel", AGENT)
                .toRole("debitorRel", AGENT).grantRole("refundBankAccount", REFERRER)

                .importEntityAlias("partnerRel", HsOfficeRelationshipEntity.class,
                        dependsOnColumn("debitorRelUuid"),
                        fetchedBySql("""
                                SELECT partnerRel.*
                                    FROM hs_office_relationship AS partnerRel
                                    JOIN hs_office_relationship AS debitorRel
                                        ON debitorRel.relType = 'ACCOUNTING' AND debitorRel.relAnchorUuid = partnerRel.relHolderUuid
                                    WHERE partnerRel.relType = 'PARTNER'
                                        AND ${REF}.debitorRelUuid = debitorRel.uuid
                                """)
                )
                .toRole("partnerRel", ADMIN).grantRole("debitorRel", ADMIN)
                .toRole("partnerRel", AGENT).grantRole("debitorRel", AGENT)
                .toRole("debitorRel", AGENT).grantRole("partnerRel", TENANT)
                .declarePlaceholderEntityAliases("partnerPerson", "operationalPerson")
                .forExampleRole("partnerPerson", ADMIN).wouldBeGrantedTo("partnerRel", ADMIN)
                .forExampleRole("operationalPerson", ADMIN).wouldBeGrantedTo("partnerRel", ADMIN)
                .forExampleRole("partnerRel", TENANT).wouldBeGrantedTo("partnerPerson", REFERRER);
    }

    public static void main(String[] args) throws IOException {
        rbac().generateWithBaseFileName("273-hs-office-debitor-rbac");
    }
}
